from CGATReport.Tracker import TrackerSQL


class ExpressionLevels(TrackerSQL):
    """Expression level measurements."""

    def __call__(self, track):
        statement = """SELECT e1.expression AS experiment1,
                              e2.expression AS experiment2,
                              e1.function as gene_function
                       FROM experiment1_data as e1,
                                 experiment2_data as e2
                       WHERE e1.gene_id = e2.gene_id
                    """

        return self.getDataFrame(statement)




import random
from CGATReport.Tracker import Tracker
from collections import OrderedDict as odict


class LongLabelsSmall(Tracker):

    """example with long labels."""
    wordsize = 5
    numwords = 5
    slices = ("small", "large", "gigantic")
    tracks = ("track1", "track2", "track3")

    def __call__(self, track, slice=None):
        if slice == "small":
            ncolumns = 10
        elif slice == "large":
            ncolumns = 40
        elif slice == "gigantic":
            ncolumns = 100

        if slice == "small":
            ncolumns = 2
        elif slice == "large":
            ncolumns = 2
        elif slice == "gigantic":
            ncolumns = 4

        data = []
        for x in range(0, ncolumns):
            label = "%s:%i %s" % (
                slice, x, " ".join(["a" * self.wordsize
                                    for y in range(self.numwords)]))
            data.append((label, random.randint(0, 100)))

        return odict(sorted(data))


class LargeMatrix(Tracker):

    """example of a large matrix with long labels."""
    wordsize = 5
    numwords = 5

    slices = ("small", "large")
    tracks = ["track%i" % i for i in range(5)]

    def __call__(self, track, slice=None):
        if slice == "small":
            ncolumns = 10
        elif slice == "large":
            ncolumns = 40

        data = []
        for x in range(0, ncolumns):
            # label = "%s:%i %s" % (slice, x, " ".join([ "a" * self.wordsize for y in range(self.numwords) ]))
            label = "%s" % (
                " ".join(["a" * self.wordsize for y in range(self.numwords)]))
            data.append((label, random.randint(0, 100)))
        return odict(data)


class LayoutTest(Tracker):

    """Layout testing."""
    nslices = 3
    nsamples = 100

    slices = ["slice%i" % i for i in range(3)]
    tracks = ["track%i" % i for i in range(5)]

    def __call__(self, track, slice=None):
        return odict((("data", [random.gauss(0, 1) for x in
                                range(self.nsamples)]),))


class SplittingTest(Tracker):

    '''return a single column of data.'''
    tracks = ["track%i" % x for x in range(0, 2)]
    slices = ["slice%i" % x for x in range(0, 10)]

    def __call__(self, track, slice=None):
        s = [random.randint(0, 20) for x in range(10)]
        random.shuffle(s)
        return odict((('x', list(range(len(s)))), ('y', s)))


class MultipleHistogramTest(Tracker):

    """Layout testing."""
    nslices = 3
    nsamples = 100

    slices = ["slice%i" % i for i in range(3)]
    tracks = ["track%i" % i for i in range(5)]

    def __call__(self, track, slice=None):
        return odict((("bin", "value-set1", "value-set2"),
                      (range(0, self.nsamples),
                       [random.gauss(0, 1) for x in range(self.nsamples)],
                       [random.gauss(0, 1) for x in range(self.nsamples)])))


class MultiLevelTable(Tracker):

    ncols = 3
    mNumLevels = 3

    tracks = ["track%i" % i for i in range(5)]

    def __call__(self, track, slice=None):
        data = [
            ["value%i" % y for y in range(self.mNumLevels)]
            for z in range(self.ncols)]

        return odict(zip(["col%i" % x for x in range(self.ncols)], data))


class LargeTable(Tracker):

    '''test case covering rendering large tables.'''

    ncols = 40
    nrows = 200

    tracks = ["track%i" % i for i in range(5)]

    def __call__(self, track, slice=None):

        data = ["value=%i" % random.randint(0, 100) for y in range(self.nrows)]
        return odict([("col%i" % x, data) for x in range(self.ncols)])


class VeryLargeMatrix(Tracker):

    """example of a large matrix with long labels."""

    # do not cache - slow in shelve as many key-value pairs
    cache = False

    ncols = 7000
    tracks = ["track%i" % i for i in range(200)]

    def __call__(self, track, slice=None):

        data = []
        for x in range(0, self.ncols):
            data.append(("col%i%s" % (x, "f" * 193), x))

        return odict(data)


class DeepTree(Tracker):

    '''example of a deeple nested data tree '''
    tracks = ("track1", "track2", "track3")
    nlevels = 8
    nslices = 2

    def __call__(self, track):

        root = odict()
        current = [root]

        for level in range(self.nlevels):
            new = []
            for x in current:
                for y in range(self.nslices):
                    o = odict()
                    x['slice%i' % y] = o
                    new.append(o)
            current = new

        for x in current:
            x['data'] = random.randint(0, 100)

        return root



from CGATReport.Tracker import Tracker

import os
import re


class WordCounter(Tracker):

    """Counting word size."""

    def getTracks(self):
        return ("all", ".py", ".rst")

    def __call__(self, track):
        word_sizes = []

        if track == "all":
            tracks = [".py", ".rst"]
        else:
            tracks = [track]

        for root, dirs, files in os.walk('.'):
            for f in files:
                fn, ext = os.path.splitext(f)
                if ext not in tracks:
                    continue
                infile = open(os.path.join(root, f), "r")
                words = re.split("\s+", "".join(infile.readlines()))
                word_sizes.extend([len(word) for word in words])
                infile.close()

        return {"word sizes": word_sizes}



import math
import numpy
import scipy
from functools import reduce

# See http://projects.scipy.org/scipy/ticket/1739
# scipy 0.11 for python3 broken, should be fixed for scipy 0.12
try:
    import scipy.stats
except ValueError:
    scipy.stats = None

try:
    from rpy2.robjects import r as R
    import rpy2.robjects.numpy2ri
except ImportError:
    R = None

from collections import OrderedDict as odict


def getSignificance(pvalue, thresholds=[0.05, 0.01, 0.001]):
    """return cartoon of significance of a p-Value."""
    n = 0
    for x in thresholds:
        if pvalue > x:
            return "*" * n
        n += 1
    return "*" * n


class Result(object):

    '''allow both member and dictionary access.'''
    slots = ("_data")

    def __init__(self):
        object.__setattr__(self, "_data", odict())

    def fromR(self, take, r_result):
        '''convert from an *r_result* dictionary using map *take*.

        *take* is a list of tuples mapping a field to the corresponding
        field in *r_result*.
        '''
        # deactivate numpy2ri conversion, interferes with .rx
        rpy2.robjects.numpy2ri.deactivate()
        for x, y in take:
            if y:
                try:
                    self._data[x] = r_result.rx(y)[0][0]
                except TypeError:
                    self._data[x] = "NA"
            else:
                self._data[x] = r_result.rx(x)[0][0]

        rpy2.robjects.numpy2ri.activate()

        return self

    def __len__(self):
        return self._data.__len__()

    def __getattr__(self, key):
        if not key.startswith("_"):
            try:
                return object.__getattribute__(self, "_data")[key]
            except KeyError:
                pass
        return getattr(self._data, key)

    def asDict(self):
        return self._data

    def keys(self):
        return list(self._data.keys())

    def values(self):
        return list(self._data.values())

    def __iter__(self):
        return self._data.__iter__()

    def __str__(self):
        return str(self._data)

    def __contains__(self, key):
        return key in self._data

    def __getitem__(self, key):
        return self._data[key]

    def __delitem__(self, key):
        del self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __setattr__(self, key, value):
        if not key.startswith("_"):
            self._data[key] = value
        else:
            object.__setattr__(self, key, value)

    def __getstate__(self):
        # required for correct pickling/unpickling
        return object.__getattribute__(self, "_data")

    def __setstate__(self, d):
        # required for correct unpickling, otherwise
        # maximum recursion threshold will be reached
        object.__setattr__(self, "_data", d)

#################################################################
#################################################################
#################################################################
# Perform log likelihood test


class LogLikelihoodTest(Result):

    def __init__(self):
        pass


def doLogLikelihoodTest(complex_ll, complex_np,
                        simple_ll, simple_np,
                        significance_threshold=0.05):
    """perform log-likelihood test between model1 and model2.
    """

    assert complex_ll >= simple_ll, "log likelihood of complex model smaller than for simple model: %f > %f" % (
        complex_ll, simple_ll)

    chi = 2 * (complex_ll - simple_ll)
    df = complex_np - simple_np

    if df <= 0:
        raise ValueError("difference of degrees of freedom not larger than 0")

    p = scipy.stats.chisqprob(chi, df)

    l = LogLikelihoodTest()

    l.mComplexLogLikelihood = complex_ll
    l.mSimpleLogLikelihood = simple_ll
    l.mComplexNumParameters = complex_np
    l.mSimpleNumParameters = simple_np
    l.mSignificanceThreshold = significance_threshold
    l.mProbability = p
    l.mChiSquaredValue = chi
    l.mDegreesFreedom = df

    if p < significance_threshold:
        l.mPassed = True
    else:
        l.mPassed = False

    return l

#################################################################
#################################################################
#################################################################


class BinomialTest:

    def __init__(self):
        pass


def doBinomialTest(p, sample_size, observed, significance_threshold=0.05):
    """perform a binomial test.

    Given are p: the probability of the NULL hypothesis, the sample_size
    and the number of observed counts.
    """
    pass

#################################################################
#################################################################
#################################################################


class ChiSquaredTest:

    def __init__(self):
        pass


def doChiSquaredTest(matrix, significance_threshold=0.05):
    """perform chi-squared test on a matrix.
    """
    nrows, ncols = matrix.shape
    if nrows != 2 or ncols != 2:
        raise ValueError(
            "chi-square currently only implemented for 2x2 tables.")

    df = (nrows - 1) * (ncols - 1)

    row_sums = [matrix[x, :].sum() for x in range(nrows)]
    col_sums = [matrix[:, x].sum() for x in range(ncols)]
    sample_size = float(sum(row_sums))

    chi = 0.0
    for x in range(nrows):
        for y in range(ncols):
            expected = row_sums[x] * col_sums[y] / sample_size
            d = matrix[x, y] - expected
            chi += d * d / expected

    result = ChiSquaredTest()

    result.mProbability = scipy.stats.chisqprob(chi, df)
    result.mDegreesFreedom = df
    result.mChiSquaredValue = chi
    result.mPassed = result.mProbability < significance_threshold
    result.mSignificance = getSignificance(result.mProbability)
    result.mSampleSize = sample_size
    result.mPhi = math.sqrt(result.mChiSquaredValue / result.mSampleSize)
    return result


def doPearsonChiSquaredTest(p, sample_size, observed,
                            significance_threshold=0.05):
    """perform a pearson chi squared test.

    Given are p: the probability of the NULL hypothesis, the sample_size
    and the number of observed counts.

    For large sample sizes, this test is a continuous approximation to
    the binomial test.
    """
    e = float(p) * sample_size
    d = float(observed) - e
    chi = d * d / e
    df = 1

    result = ChiSquaredTest()

    result.mProbability = scipy.stats.chisqprob(chi, df)
    result.mDegreesFreedom = df
    result.mChiSquaredValue = chi
    result.mPassed = result.mProbability < significance_threshold
    result.mSignificance = getSignificance(result.mProbability)
    result.mSampleSize = sample_size
    result.mPhi = math.sqrt(result.mChiSquaredValue / result.mSampleSize)
    result.mObserved = observed
    result.mExpected = e
    return result

#################################################################
#################################################################
#################################################################
# Convenience functions and objects for statistical analysis


class Summary(Result):

    """a collection of distributional parameters. Available properties
    are:

    mean, median, min, max, samplestd, sum, counts
    """

    def __init__(self, values=None, format="%6.4f", mode="float"):
        Result.__init__(self)
        self._format = format
        self._mode = mode
        # note that this determintes the order of the fields at output
        self.counts, self.min, self.max, self.mean, self.median, self.samplestd, self.sum, self.q1, self.q3 = \
            (0, 0, 0, 0, 0, 0, 0, 0, 0)

        if values is not None:

            values = [x for x in values if x != None]

            if len(values) == 0:
                raise ValueError("no data for statistics")

            # convert
            self._nerrors = 0
            if type(values[0]) not in (int, float):
                n = []
                for x in values:
                    try:
                        n.append(float(x))
                    except ValueError:
                        self._nerrors += 1
            else:
                n = values

            if len(n) == 0:
                raise ValueError("no data for statistics")

            # use a non-sort algorithm later.
            n.sort()
            self.q1 = n[len(n) / 4]
            self.q3 = n[len(n) * 3 / 4]

            self.counts = len(n)
            self.min = min(n)
            self.max = max(n)
            self.mean = numpy.mean(n)
            self.median = numpy.median(n)
            self.samplestd = numpy.std(n)
            self.sum = reduce(lambda x, y: x + y, n)

    def getHeaders(self):
        """returns header of column separated values."""
        return ("nval", "min", "max", "mean", "median", "stddev", "sum", "q1", "q3")

    def getHeader(self):
        """returns header of column separated values."""
        return "\t".join(self.getHeaders())

    def __str__(self):
        """return string representation of data."""

        if self._mode == "int":
            format_vals = "%i"
            format_median = "%.1f"
        else:
            format_vals = self._format
            format_median = self._format

        return "\t".join(("%i" % self.counts,
                          format_vals % self.min,
                          format_vals % self.max,
                          self._format % self.mean,
                          format_median % self.median,
                          self._format % self.samplestd,
                          format_vals % self.sum,
                          format_vals % self.q1,
                          format_vals % self.q3,
                          ))


class FDRResult:

    def __init__(self):
        pass

    def plot(self, hardcopy=None):

        if hardcopy:
            R.png(hardcopy, width=1024, height=768, type="cairo")

        R.require('qvalue')

        # build a qobj
        R.assign("pval", self.mPValues)
        R.assign("pi0", self.mPi0)
        R.assign("qval", self.mQValues)
        R.assign("lambda", self.mLambda)
        R("""qobj <-list(pi0=pi0, qvalues=qval, pvalues=pval, lambda=lambda)""")
        R(""" class(qobj) <- "qvalue" """)

        R("""qplot(qobj)""")

        if hardcopy:
            R.dev_off()


def doFDR(pvalues,
          vlambda=numpy.arange(0, 0.95, 0.05),
          pi0_method="smoother",
          fdr_level=None,
          robust=False,
          smooth_df=3,
          smooth_log_pi0=False):
    """modeled after code taken from http://genomics.princeton.edu/storeylab/qvalue/linux.html.

    I did not like the error handling so I translated most to python.
    """

    if min(pvalues) < 0 or max(pvalues) > 1:
        raise ValueError("p-values out of range")

    if len(vlambda) > 1 and len(vlambda) < 4:
        raise ValueError(
            " If length of vlambda greater than 1, you need at least 4 values.")

    if len(vlambda) > 1 and (min(vlambda) < 0 or max(vlambda) >= 1):
        raise ValueError("vlambda must be within [0, 1).")

    m = len(pvalues)

    # these next few functions are the various ways to estimate pi0
    if len(vlambda) == 1:
        vlambda = vlambda[0]
        if vlambda < 0 or vlambda >= 1:
            raise ValueError("vlambda must be within [0, 1).")

        pi0 = numpy.mean([x >= vlambda for x in pvalues]) / (1.0 - vlambda)
        pi0 = min(pi0, 1.0)
        R.assign("pi0", pi0)
    else:
        pi0 = numpy.zeros(len(vlambda), numpy.float)

        for i in range(len(vlambda)):
            pi0[i] = numpy.mean([x >= vlambda[i]
                                 for x in pvalues]) / (1.0 - vlambda[i])

        R.assign("pi0", pi0)
        R.assign("vlambda", vlambda)

        if pi0_method == "smoother":
            if smooth_log_pi0:
                pi0 = math.log(pi0)

            R.assign("smooth_df", smooth_df)

            spi0 = R("""spi0 <- smooth.spline(vlambda,pi0, df = smooth_df)""")
            pi0 = R("""pi0 <- predict(spi0, x = max(vlambda))$y""")

            if smooth_log_pi0:
                pi0 = math.exp(pi0)

        elif pi0_method == "bootstrap":

            minpi0 = min(pi0)

            mse = numpy.zeros(len(vlambda), numpy.float)
            pi0_boot = numpy.zeros(len(vlambda), numpy.float)

            R.assign("pvalues", pvalues)
            pi0 = R("""
            m <- length(pvalues)
            minpi0 <- min(pi0)
            mse <- rep(0,length(vlambda))
            pi0_boot <- rep(0,length(vlambda))
            for(i in 1:100)
            {
                pvalues_boot <- sample(pvalues,size=m,replace=TRUE)
                for(i in 1:length(vlambda))
                {
                    pi0_boot[i] <- mean(pvalues_boot>vlambda[i])/(1-vlambda[i])
                }
                mse <- mse + (pi0_boot-minpi0)^2
            }
            pi0 <- min(pi0[mse==min(mse)])""")
        else:
            raise ValueError(
                "'pi0_method' must be one of 'smoother' or 'bootstrap'.")

        pi0 = min(pi0, 1.0)

    if pi0 <= 0:
        raise ValueError(
            "The estimated pi0 <= 0. Check that you have valid p-values or use another vlambda method.")

    if fdr_level != None and (fdr_level <= 0 or fdr_level > 1):
        raise ValueError("'fdr_level' must be within (0, 1].")

    # The estimated q-values calculated here
    #u = numpy.argsort(p)

    # change by Alan
    # ranking function which returns number of observations less than or equal
    R.assign("pvalues", pvalues)
    R.assign("robust", robust)
    qvalues = R("""u <- order(pvalues)
    qvalues.rank <- function(x)
{
      idx <- sort.list(x)

      fc <- factor(x)
      nl <- length(levels(fc))
      bin <- as.integer(fc)
      tbl <- tabulate(bin)
      cs <- cumsum(tbl)

      tbl <- rep(cs, tbl)
      tbl[idx] <- tbl

      return(tbl)
}

v <- qvalues.rank(pvalues)
m <- length(pvalues)

qvalues <- pi0 * m * pvalues / v
if(robust)
{
        qvalues <- pi0*m*pvalues/(v*(1-(1-pvalues)^m))
}
qvalues[u[m]] <- min(qvalues[u[m]],1)
for(i in (m-1):1)
{
   qvalues[u[i]] <- min(qvalues[u[i]],qvalues[u[i+1]],1)
}
qvalues
""")

    result = FDRResult()
    result.mQValues = qvalues

    if fdr_level != None:
        result.mPassed = [x <= fdr_level for x in result.mQValues]

    result.mPValues = pvalues
    result.mPi0 = pi0
    result.mLambda = vlambda

    return result

#################################################################
#################################################################
#################################################################


class CorrelationTest(Result):

    def __init__(self,
                 r_result=None,
                 s_result=None,
                 method=None,
                 nobservations=0):
        Result.__init__(self)

        self.pvalue = None
        self.method = method
        self.nobservations = 0

        if r_result:
            self.coefficient = r_result['estimate']['cor']
            self.pvalue = float(r_result['p.value'])
            self.nobservations = r_result['parameter']['df']
            self.method = r_result['method']
            self.alternative = r_result['alternative']
        elif s_result:
            self.coefficient = s_result[0]
            self.pvalue = s_result[1]
            self.nobservations = nobservations
            self.alternative = "two-sided"

        if self.pvalue != None:
            if self.pvalue > 0:
                self.logpvalue = math.log(self.pvalue)
            else:
                self.logpvalue = 0
            self.significance = getSignificance(self.pvalue)

    def __str__(self):
        return "\t".join((
            "%6.4f" % self.coefficient,
            "%e" % self.pvalue,
            self.significance,
            "%i" % self.nobservations,
            self.method,
            self.alternative))


def filterMasked(xvals, yvals, missing=("na", "Nan", None, ""), dtype = numpy.float):
    """convert xvals and yvals to numpy array skipping pairs with
    one or more missing values."""
    xmask = [i in missing for i in xvals]
    ymask = [i in missing for i in yvals]
    return (numpy.array([xvals[i] for i in range(len(xvals)) if not xmask[i]], dtype=dtype),
            numpy.array([yvals[i] for i in range(len(yvals)) if not ymask[i]], dtype=dtype))


def filterNone(args, missing=("na", "Nan", None, "", 'None', 'none'), dtype = numpy.float):
    '''convert arrays in 'args' to numpy arrays of 'dtype', skipping where any of
    the columns have a value of missing.

    >>> Stats.filterNone(((1,2,3), (4,5,6)))
    [array([ 1.,  2.,  3.]), array([ 4.,  5.,  6.])]
    >>> Stats.filterNone(((1,2,3), (4,None,6)))
    [array([ 1.,  3.]), array([ 4.,  6.])]
    >>> Stats.filterNone(((None,2,3), (4,None,6)))
    [array([ 3.]), array([ 6.])]
    '''
    mi = min([len(x) for x in args])
    ma = max([len(x) for x in args])
    assert mi == ma, "arrays have unequal length to start with: min=%i, max=%i." % (
        mi, ma)

    mask = [sum([z in missing for z in x]) for x in zip(*args)]

    return [numpy.array([x[i] for i in range(len(x)) if not mask[i]], dtype=dtype) for x in args]


def filterMissing(args,
                  missing=("na", "Nan", None, "", 'None', 'none'),
                  dtype = numpy.float):
    '''remove rows in args where at least one of the columns have a
       missing value.'''

    mi = min([len(x) for x in args])
    ma = max([len(x) for x in args])
    assert mi == ma, "arrays have unequal length to start with: min=%i, max=%i." % (
        mi, ma)

    keep = numpy.array([True] * ma)
    for values in args:
        keep &= values.notnull()

    return [x[keep] for x in args]


def doCorrelationTest(xvals, yvals, method="pearson"):
    """compute correlation between x and y.

    Raises a value-error if there are not enough observations.
    """

    if scipy.stats is None:
        raise ImportError("scipy.stats not available")

    if len(xvals) <= 1 or len(yvals) <= 1:
        raise ValueError("can not compute correlation with no data")
    if len(xvals) != len(yvals):
        raise ValueError("data vectors have unequal length")

    x, y = filterMasked(xvals, yvals)

    if method == "pearson":
        s_result = scipy.stats.pearsonr(x, y)
    elif method == "spearman":
        s_result = scipy.stats.spearmanr(x, y)
    else:
        raise ValueError("unknown method %s" % (method))

    result = CorrelationTest(s_result=s_result,
                             method=method,
                             nobservations=len(x))

    return result.asDict()


###################################################################
###################################################################
###################################################################
# compute ROC curves from sorted values
###################################################################
def computeROC(values):
    '''return a roc curve for *values*.
    Values is a sorted list of (value, bool) pairs.

    '''
    roc = []

    npositives = len([x for x in values if x[1]])
    if npositives == 0:
        raise ValueError("no positives among values")

    ntotal = len(values)

    last_value, last_fpr = None, None
    tp, fp = 0, 0
    tn, fn = ntotal - npositives, npositives

    for value, is_positive in values:
        if is_positive:
            tp += 1
            fn -= 1
        else:
            fp += 1
            tn -= 1

        if last_value != value:

            try:
                tpr = float(tp) / (tp + fn)
            except ZeroDivisionError:
                tpr = 0

            try:
                fpr = float(fp) / (fp + tn)
            except ZeroDivisionError:
                fpr = 0

            if last_fpr != fpr:
                roc.append((fpr, tpr))
                last_fpr = fpr

        last_values = value

    return roc

###################################################################
###################################################################
###################################################################
##
###################################################################


def getAreaUnderCurve(xvalues, yvalues):
    '''compute area under curve from a set of discrete x,y coordinates
    using trapezoids.

    This is only as accurate as the density of points.
    '''

    assert len(xvalues) == len(yvalues)
    last_x, last_y = xvalues[0], yvalues[0]
    auc = 0
    for x, y in zip(xvalues, yvalues)[1:]:
        dx = x - last_x
        assert not dx < 0, "x not increasing: %f >= %f" % (last_x, x)
        dy = abs(last_y - y)
        my = min(last_y, y)
        # rectangle plus triangle
        auc += dx * my + dx * dy / 2
        last_x, last_y = x, y

    return auc

###################################################################
###################################################################
###################################################################
##
###################################################################


def getSensitivityRecall(values):
    '''return sensitivity/selectivity.

    Values is a sorted list of (value, bool) pairs.
    '''

    npositives = 0.0
    npredicted = 0.0
    l = None
    result = []
    total = float(len(values))
    for value, is_positive in values:
        npredicted += 1.0
        if is_positive > 0:
            npositives += 1.0
        if value != l:
            result.append((value, npositives / npredicted, npredicted / total))
        l = value
    if l:
        result.append((l, npositives / npredicted, npredicted / total))

    return result.asDict()

###################################################################
###################################################################
###################################################################
##
###################################################################


def doMannWhitneyUTest(xvals, yvals):
    '''apply the Mann-Whitney U test to test for the difference of medians.'''
    if len(xvals) == 0 or len(yvals) == 0:
        result = Result()
    else:
        if R:
            r_result = R['wilcox.test'](xvals, yvals, paired=False)
            result = Result().fromR(
                (("pvalue", 'p.value'),
                 ('alternative', None),
                 ('method', None)),
                r_result)
        else:
            raise ValueError("rpy2 not available")

    result.xobservations = len(xvals)
    result.yobservations = len(yvals)

    return result.asDict()


def buildMatrixFromEdges(edges,
                         in_map_token2row={},
                         in_map_token2col={},
                         is_square=True,
                         is_symmetric=False,
                         missing_value=0,
                         diagonal_value=0,
                         dtype=numpy.int):
    """build a matrix from an edge-list representation.

    For example, the following list of tuples::

       [('A', 'B', 1),
        ('A', 'C', 2),
        ('B', 'C', 3)]

    will be converted to the following matrix::

         A B C
       A   1 2
       B     3
       C

    If *is_symmetric* is set to True, the matrix is assumed to be
    symmetric and missing values will automatically be filled::

         A B C
       A   1 2
       B 1   3
       C 2 3

    If edge list may contain four elements, in which case the
    fourth element is expected to be the value of the lower
    diagonal in a symmetric matrix::

       [('A', 'B', 1, 4),
        ('A', 'C', 2, 5),
        ('B', 'C', 3, 6)]

    will yield::

         A B C
       A   1 2
       B 4   3
       C 5 6

    If *is_square* the matrix will be squared.

    returns a numpy matrix and lists of row and column names.
    """

    if in_map_token2row:
        map_token2row = in_map_token2row
    else:
        map_token2row = {}

    if in_map_token2col:
        map_token2col = in_map_token2col
    else:
        map_token2col = {}

    has_row_names = len(map_token2row) > 0
    has_col_names = len(map_token2col) > 0

    # if either row/column names are not given:
    if not map_token2row or not map_token2col:

        row_tokens = sorted(list(set([x[0] for x in edges])))
        col_tokens = sorted(list(set([x[1] for x in edges])))

        if not has_row_names:
            for row_token in row_tokens:
                if row_token not in map_token2row:
                    map_token2row[row_token] = len(map_token2row)
        if not has_col_names:
            for col_token in col_tokens:
                if col_token not in map_token2col:
                    map_token2col[col_token] = len(map_token2col)

        # for square matrices merge row and column labels
        if is_square:
            for col_token in map_token2col.keys():
                if col_token not in map_token2row:
                    map_token2row[col_token] = len(map_token2row)
            map_token2col = map_token2row

    matrix = numpy.matrix([missing_value] *
                          len(map_token2row) * len(map_token2col),
                          dtype=dtype)
    matrix.shape = (len(map_token2row), len(map_token2col))

    if is_square:
        for i in range(len(map_token2col)):
            matrix[i, i] = diagonal_value

    if len(edges[0]) == 3:
        if is_symmetric:
            for row, col, value in edges:
                if value is not None:
                    matrix[map_token2row[row], map_token2col[col]] = \
                        matrix[map_token2row[col], map_token2col[row]] = value
        else:
            for row, col, value in edges:
                if value is not None:
                    matrix[map_token2row[row], map_token2col[col]] = value
    elif len(edges[0]) == 4:
        for row, col, value1, value2 in edges:
            if value1 is not None:
                matrix[map_token2row[row], map_token2col[col]] = value1
            if value2 is not None:
                matrix[map_token2row[col], map_token2col[row]] = value2
    else:
        raise ValueError(
            "unexpected number of elements in list, expected 3 or 4, "
            "got %i" % (len(edges[0])))

    col_tokens = map_token2col.items()
    col_tokens.sort(lambda x, y: cmp(x[1], y[1]))
    row_tokens = map_token2row.items()
    row_tokens.sort(lambda x, y: cmp(x[1], y[1]))

    return matrix, [x[0] for x in row_tokens], [x[0] for x in col_tokens]


# taken from http://scipy-cookbook.readthedocs.io/items/SignalSmooth.html
def smooth(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError, "smooth only accepts 1 dimension arrays."

    if x.size < window_len:
        raise ValueError, "Input vector needs to be bigger than window size."

    if window_len<3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"

    s = numpy.r_[x[window_len-1:0:-1], x, x[-1:-window_len:-1]]

    if window == 'flat': #moving average
        w = numpy.ones(window_len, 'd')
    else:
        w = eval('numpy.' + window + '(window_len)')

    y = numpy.convolve(w / w.sum(), s, mode='valid')

    return y



#!/bin/env python

import web, os

from SphinxReport import Cache
from SphinxReport import Utils
from SphinxReport import DataTree

urls = ( '/data/(.*)', 'DataTable',
         '/index/(.*)', 'Index'  )

# expose zip within templates
global_vars = {'zip': zip }

render = web.template.render('templates/', globals = global_vars)

app = web.application(urls, globals() )

class DataTable:
    '''render data retrieved from cache as a table.'''

    def GET(self, tracker):

        cache = Cache.Cache( tracker, mode = "r" )
        data = DataTree.fromCache( cache )
        table, row_headers, col_headers = DataTree.tree2table( data )

        return render.data_table(table, row_headers, col_headers )


if __name__ == "__main__": app.run()






# -*- coding: utf-8 -*-
"""
    sphinx.ext.errors
    ~~~~~~~~~~~~~~~~~

    Collect all errors

:copyright: Copyright 2007-2009 by the Andreas Heger, see AUTHORS.
:license: BSD, see LICENSE for details.
"""

from docutils import nodes
import logging

from sphinx.locale import _
from sphinx.environment import NoUri
from sphinx.util.compat import Directive
from docutils.parsers.rst.directives.admonitions import BaseAdmonition


class cgatreporterror_node(nodes.warning, nodes.Element):
    pass


class cgatreporterrorlist(nodes.General, nodes.Element):
    pass


class CGATReportError(BaseAdmonition):

    node_class = cgatreporterror_node

    # accept error as an optional argument
    optional_arguments = 1

    def run(self):

        r = BaseAdmonition.run(self)

        if len(self.arguments) > 0:
            errorclass = self.arguments[0]
        else:
            errorclass = "generic"

        env = self.state.document.settings.env
        if not hasattr(env, 'cgatreporterror_all_cgatreporterrors'):
            env.cgatreporterror_all_cgatreporterrors = []

        env.cgatreporterror_all_cgatreporterrors.append({
            'docname': env.docname,
            'lineno': self.lineno,
            'cgatreporterror': r[0].deepcopy(),
            'errorclass': errorclass,
        })

        logging.error("CGATReport-Warning: %s" % errorclass)

        return r


class CGATReportErrorList(Directive):

    """
    A list of all cgatreporterror entries.
    """

    has_content = False
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {}

    def run(self):
        # Simply insert an empty cgatreporterrorlist node which will
        # be replaced later when process_cgatreporterror_nodes is
        # called
        return [cgatreporterrorlist('')]


def process_cgatreporterror_nodes(app, doctree, fromdocname):
    if not app.config['cgatreport_show_errors']:
        for node in doctree.traverse(cgatreporterror_node):
            node.parent.remove(node)

    # Replace all cgatreporterrorlist nodes with a list of the
    # collected cgatreporterrors.  Augment each cgatreporterror with a
    # backlink to the original location.
    env = app.builder.env

    if not hasattr(env, 'cgatreporterror_all_cgatreporterrors'):
        env.cgatreporterror_all_cgatreporterrors = []

    for node in doctree.traverse(cgatreporterrorlist):
        if not app.config['cgatreport_show_errors']:
            node.replace_self([])
            continue

        content = []
        nerrors = 0

        para = nodes.paragraph()
        para += nodes.Text("There are %i errors" %
                           len(env.cgatreporterror_all_cgatreporterrors))
        content.append(para)

        for cgatreporterror_info in env.cgatreporterror_all_cgatreporterrors:

            para = nodes.paragraph()

            filename = env.doc2path(
                cgatreporterror_info['docname'], base=None)

            nerrors += 1
            location_str = '%s:%d ' % (
                filename, cgatreporterror_info['lineno'])
            try:
                description_str = cgatreporterror_info['errorclass']
            except KeyError:
                description_str = "unknown"

            # Create a reference
            newnode = nodes.reference('', '')
            innernode = nodes.emphasis(_(location_str), _(location_str))
            newnode['refdocname'] = cgatreporterror_info['docname']
            try:
                newnode['refuri'] = app.builder.get_relative_uri(
                    fromdocname, cgatreporterror_info['docname'])
            except NoUri:
                # ignore if no URI can be determined, e.g. for LaTeX output
                pass
            newnode.append(innernode)

            para += newnode
            para += nodes.Text(description_str, description_str)
            para += nodes.Text("\n", "\n")

            # (Recursively) resolve references in the cgatreporterror content
            cgatreporterror_entry = cgatreporterror_info[
                'cgatreporterror']
            env.resolve_references(cgatreporterror_entry,
                                   cgatreporterror_info['docname'],
                                   app.builder)

            content.append(para)

        node.replace_self(content)


def purge_cgatreporterrors(app, env, docname):
    if not hasattr(env, 'cgatreporterror_all_cgatreporterrors'):
        return
    env.cgatreporterror_all_cgatreporterrors = \
        [cgatreporterror for cgatreporterror in
         env.cgatreporterror_all_cgatreporterrors
         if cgatreporterror['docname'] != docname]


def visit_cgatreporterror_node(self, node):
    self.visit_admonition(node)


def depart_cgatreporterror_node(self, node):
    self.depart_admonition(node)


def setup(app):
    app.add_config_value('cgatreport_show_errors', True, False)

    app.add_node(cgatreporterrorlist)
    app.add_node(
        cgatreporterror_node,
        html=(
            visit_cgatreporterror_node, depart_cgatreporterror_node),
        latex=(
            visit_cgatreporterror_node,
            depart_cgatreporterror_node),
        text=(visit_cgatreporterror_node, depart_cgatreporterror_node))

    app.add_directive('error', CGATReportError)
    app.add_directive('errorlist', CGATReportErrorList)
    app.connect('doctree-resolved', process_cgatreporterror_nodes)
    app.connect('env-purge-doc', purge_cgatreporterrors)



import os
import re
from CGATReport.Component import Component


class RSTPlugin(Component):

    '''collect rst text.

    This plugin looks for image/figure directives in
    literal text output and changes their path so that
    they are valid within sphinx.

    images: path relative to rst directory

    links: path relative to html content

    '''

    capabilities = ['collect']

    # include white spaces at the end for length-neutral substitution
    # Pattern will also include a terminating '"'. This is used to
    # test if the context is a csv table, in which case no padding
    # should be done.
    rx_img = re.compile("\.\. (image|figure):: ([^ |+,:]+[ ]*)")

    # external link targets:
    # .. _link: target
    rx_link = re.compile("\.\. _([^ |+,:]+)\s*:\s*([^ |+,:]+)[ ]*")

    def __init__(self, *args, **kwargs):
        Component.__init__(self, *args, **kwargs)

    def collect(self,
                blocks,
                template_name,
                outdir,
                rstdir,
                builddir,
                srcdir,
                content,
                display_options,
                tracker_id,
                links={}):
        '''collect rst output from result blocks.

        '''
        map_figure2text = {}

        def replace_and_pad(s, old, new):
            '''replace old with new in s such that length of
            new+spaces is at least that of old+spaces.

            There needs to be enough space for padding.
            '''
            # do not pad in csv tables
            if not old.endswith('"'):
                oldlen = len(old)
                newlen = len(new)
                new = new + " " * (oldlen - newlen)
            n = s.replace(old, new)
            return n

        for xblocks in blocks:
            for block in xblocks:
                if not hasattr(block, "text"):
                    continue
                lines = block.text.split("\n")
                n = []
                for l in lines:
                    ll = l
                    for x in self.rx_img.finditer(ll):
                        directive, filename = x.groups()
                        relpath = os.path.relpath(
                            filename.strip(),
                            os.path.abspath(rstdir))
                        newpath = re.sub("\\\\", "/", relpath)
                        l = replace_and_pad(l, filename, newpath)

                    for x in self.rx_link.finditer(ll):
                        directive, filename = x.groups()
                        newpath = re.sub(
                            "\\\\", "/",
                            os.path.abspath(filename.strip()))
                        # pad with spaces to keep table alignment
                        l = replace_and_pad(l, filename, newpath)

                    n.append(l)
                block.text = "\n".join(n)

        return map_figure2text



#!/bin/env python

"""
cgatreport-clean
==================

:command:`cgatreport-clean` removes all documents associated
with:class:`Tracker` thus allowing it to be re-built the next
time:command:`sphinx` is invoked as::

   cgatreport-clean [clean|distclean|cache|Tracker1] [Tracker2] [...]

The full list of command line options is listed by suppling:option:`-h/--help`
on the command line.

The options are:

**-v/--verbose** verbosity level
   Increase the number of status messages displayed.

**-n/--dry-run**
   Show all files that will be removed but do not remove them.

**-s/--section** choice
   Only clean within certain types of documents. The default is all. Possible choices are
   ``tracker`` and ``text``.

**-b/--build** path
   Path to build directory. By default, documents are examined in the current directory.


**-w/--path** path
   Path with restructured document. By default, the ``.rst`` files are assumed to reside
   in the current directory. Without the documents, the clean command will not be able
   to remove all documents that refer to a:term:`tracker`.

If there is only one target and it is ``clean``, ``distclean``,
the full build we cleaned up. If it is ``cache``, only the cache
will be cleaned forcing newly built:class:`Tracker` objects to recompute
their data.

Alternatively, if one or more than one:class:`Tracker` is given, all
documents referencing these will be removed to force a re-built next
time:command:`sphinx` is invoked.

Trackers are interpreted as patterns. For example,
the following will remove all data from the cache and all previously build
documents containing trackers matching the word ``OldData``::

   cgatreport-clean OldData

"""

import sys
import os
import imp
import io
import re
import types
import glob
import optparse
import shutil

USAGE = """python %s [OPTIONS] target

clean all data according to target.

Targets can contain wild cards.

""" % sys.argv[0]

from CGATReport.Tracker import Tracker

SEPARATOR = "@"

# import conf.py for source_suffix
source_suffix = ".rst"
if os.path.exists("conf.py"):
    try:
        exec(compile(open("conf.py").read(), "conf.py", 'exec'))
    except ValueError:
        pass

RSTDIR = "."
if "docsdir" in locals():
    RSTDIR = docsdir


def deleteFiles(test_f, dirs_to_check=(".",), dry_run = False):
    """remove all files that test_f returns True for.
    """
    removed = []
    for d in dirs_to_check:
        for root, dirs, files in os.walk(d):
            for f in files:
                if test_f(f):
                    try:
                        ff = os.path.join(root, f)
                        if not dry_run:
                            os.remove(ff)
                        removed.append(ff)
                    except OSError as msg:
                        pass

    return removed


def removeTracker(tracker,
                  dry_run=False,
                  builddir="report"):
    """remove all files created by:class:Renderer objects
    that use tracker.
    """
    # get locations
    # this is a patch - add configuration options from conf.py
    dirs_to_check = ("_static", "_cache", "_build", builddir)

    pattern = ".*%s.*" % tracker
    # image and text files
    rx1 = re.compile("_%s%s" % (pattern, SEPARATOR))
    # files in cache
    rx2 = re.compile("_%s$" % (pattern))
    # .code files
    rx3 = re.compile("_%s%s" % (pattern, ".code"))
    # .html files
    rx4 = re.compile("_%s%s" % (pattern, ".html"))
    test_f = lambda x: rx1.search(x) or rx2.search(
        x) or rx3.search(x) or rx4.search(x)

    return deleteFiles(test_f, dirs_to_check, dry_run=dry_run)


def removeText(pattern,
               dry_run=False,
               sourcedir=".",
               builddir="report",
               suffix=".rst"):
    """remove all files that contain the ``pattern``."""

    # find all .rst files that reference tracker
    nremoved = 0
    rx_tracker = re.compile(pattern)
    files_to_check = []
    for root, dirs, files in os.walk(sourcedir):
        for f in files:
            if f.endswith(suffix):
                fn = os.path.join(root, f)
                try:
                    data = "".join(open(fn, "r").readlines())
                except IOError:
                    continue

                found = rx_tracker.search(data)

                if found:
                    files_to_check.append(f)

    suffixes = (".doctree", ".html")
    patterns = []
    for f in files_to_check:
        p = f[:-len(suffix)]
        for s in suffixes:
            patterns.append(re.compile("%s%s$" % (p, s)))

    def test_f(x):
        for p in patterns:
            if p.search(x):
                return True
        return False

    dirs_to_check = (builddir,)

    return deleteFiles(test_f, dirs_to_check, dry_run=dry_run)


def main():

    parser = optparse.OptionParser(version="%prog version: $Id$", usage=USAGE)

    parser.add_option("-v", "--verbose", dest="loglevel", type="int",
                      help="loglevel. The higher, the more output [default=%default]")

    parser.add_option("-s", "--section", dest="sections", type="choice", action="append",
                      choices=("tracker", "text", "doctree"),
                      help="only clean from certain sections [default=%default]")

    parser.add_option("-p", "--path", dest="path", type="string",
                      help="path to rst source [default=%default]")

    parser.add_option("-b", "--build", dest="builddir", type="string",
                      help="path to build dir [default=%default]")

    parser.add_option("-n", "--dry-run", dest="dry_run", action="store_true",
                      help="only show what is about to be deleted, but do not delete [default=%default]")

    parser.set_defaults(loglevel=2,
                        dry_run=False,
                        path=RSTDIR,
                        builddir=".",
                        sections=[])

    (options, args) = parser.parse_args()

    if len(args) == 0:
        print(USAGE)
        raise ValueError("please supply at least one target.""")

    if len(args) == 1 and args[0] in ("clean", "distclean", "cache"):
        dirs = []
        target = args[0]
        if target in ("clean", "distclean"):
            dirs.append("_build")

        if target in ("cache", "distclean"):
            dirs.append("_cache")

        if target in ("distclean",):
            dirs.append("_static/report_directive")

        if options.dry_run:
            print("the following directories will be deleted:")
            print("\n".join(dirs))
        else:
            for d in dirs:
                if os.path.exists(d):
                    shutil.rmtree(d)

    else:
        if options.dry_run:
            print("the following files will be deleted:")

        for tracker in args:
            print("cleaning up %s ..." % tracker)
            removed = []
            if not options.sections or "tracker" in options.sections:
                if options.loglevel >= 2:
                    print("removing trackers")
                removed.extend(removeTracker(tracker, dry_run=options.dry_run))
            if not options.sections or "doctree" in options.sections:
                if options.loglevel >= 2:
                    print("removing doctrees")
                removed.extend(removeText(tracker,
                                          dry_run=options.dry_run,
                                          sourcedir=options.builddir,
                                          builddir=options.builddir,
                                          suffix=".doctree"))
            if not options.sections or "text" in options.sections:
                if options.loglevel >= 2:
                    print("removing rst")
                if options.loglevel >= 2:
                    print("sourcedir=%s" % options.path)
                removed.extend(removeText(tracker,
                                          dry_run=options.dry_run,
                                          sourcedir=options.path,
                                          builddir=options.builddir,
                                          suffix=source_suffix))
            print("%i files (done)" % len(removed))
            if options.loglevel >= 3:
                print("\n".join(removed))
            if options.dry_run:
                print("\n".join(removed))
if __name__ == "__main__":
    sys.exit(main())



from CGATReport.Tracker import Tracker
from collections import OrderedDict as odict
import matplotlib.dates
import datetime


class ProjectDatesExample(Tracker):
    tracks = ("proj1", "proj2", "proj3")

    def __call__(self, track):

        # define a convenience function to convert
        # a three-number tuple to a scalar date float:
        f = lambda year, month, day: matplotlib.dates.date2num(
            datetime.datetime(year, month, day))

        if track == "proj1":
            return odict((("start", f(2012, 1, 1)),
                          ("duration", 100)))
        elif track == "proj2":
            return odict((("start", f(2012, 6, 1)),
                          ("duration", 200)))
        elif track == "proj3":
            return odict((("start", f(2012, 8, 1)),
                          ("duration", 100)))



'''Plotting using the ggplot module.'''

from CGATReportPlugins.Renderer import Renderer
from CGATReportPlugins.Plotter import Plotter
from docutils.parsers.rst import directives

import matplotlib.pyplot as plt
import pandas


class PandasPlot(Renderer, Plotter):

    """Use the python ggplot libary for plotting.
    """
    options = (
        ('statement',  directives.unchanged),
    ) + Renderer.options + Plotter.options

    group_level = 0

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

        self.statement = kwargs.get('statement')

    def render(self, dataframe, path):

        # the index in the dataframe is reset in order
        # to add the index as a column.

        # Currently, ggplot is not using hierarchical indices, but
        # see this thread: https://github.com/yhat/ggplot/issues/285

        dataframe.reset_index(inplace=True)

        p = None
        s = "p = dataframe.plot(%s)" % self.statement

        try:
            exec s in globals(), locals()
        except Exception, msg:
            raise Exception(
                "pandas.plot() raised error for statement '%s': msg=%s" %
                (s, msg))

        self.mFigure += 1
        if self.title:
            plt.title(self.title)
        plts = [p]
        return self.endPlot(plts, None, path)



'''Configuration values in CGATReport.

'''

import os

SEPARATOR = "@"

HTML_IMAGE_FORMAT = ('main', 'png', 80)
LATEX_IMAGE_FORMAT = ()  # ('pdf', 'pdf' 50)

ADDITIONAL_FORMATS = [
    ('hires', 'hires.png', 200),
    # ('eps', 'eps', 50),
    # ('svg', 'svg', 50),
    # ('pdf', 'pdf', 50),
]



#!/usr/bin/env python
'''unit testing code for SphinxReport
'''

import unittest
import sys

IS_PYTHON3 = sys.version_info[0] >= 3

import SphinxReport
import SphinxReport.Utils
import SphinxReport.Dispatcher


class SphinxReportTest(unittest.TestCase):
    ''' '''
    renderer = "table"
    tracker = "tests.TestTrackers.LabeledDataExample"

    def testTracker(self):
        code, tracker, tracker_path = SphinxReport.Utils.makeTracker(
            self.tracker, (), {})
        renderer = SphinxReport.Utils.getRenderer(
            self.renderer, {})

        dispatcher = SphinxReport.Dispatcher.Dispatcher(
            tracker,
            renderer,
            [])
        results = dispatcher({})
        print (results)
        return True




from CGATReport.Tracker import Tracker
from CGATReport.DataTypes import returnLabeledData, returnSingleColumnData, returnMultipleColumnData, returnMultipleColumns


class SingleColumnDataExample(Tracker):
    mData = (1, 2, 3)

    def getTracks(self):
        return "track1", "track2"

    @returnSingleColumnData
    def __call__(self, track, slice=None):
        if track == "track1":
            return self.mData * 1
        elif track == "track2":
            return self.mData * 2


class SingleColumnDataExample2(SingleColumnDataExample):
    mData = (10, 20, 30)


@returnSingleColumnData
def FunctionExample1():
    return (1, 2, 3)


@returnSingleColumnData
def FunctionExample2():
    return (4, 5, 6)

if __name__ == "__main__":

    print (list(SingleColumnDataExample()("track1")))
    print (list(SingleColumnDataExample()("track2")))
    print (list(SingleColumnDataExample()("track1")))

    a = SingleColumnDataExample()
    print (list(a("track1")))
    print (list(a("track2")))
    print (list(a("track1")))

    print (list(SingleColumnDataExample2()("track1")))
    print (list(SingleColumnDataExample2()("track2")))
    print (list(SingleColumnDataExample2()("track1")))

    a = SingleColumnDataExample2()
    print (list(a("track1")))
    print (list(a("track2")))
    print (list(a("track1")))

    print (list(FunctionExample1()))

    print (list(FunctionExample2()))



import numpy
import numpy.linalg
import numpy.linalg.linalg


def GetIndices(matrix):
    """return order (1st eigenvector) of row and column indicies.

    """

    original_nrows, original_ncols = matrix.shape
    # calculate row and column sums
    row_sums = numpy.array(numpy.sum(matrix, 1)).flatten()
    col_sums = numpy.array(numpy.sum(matrix, 0)).flatten()

    # check for empty rows/columns
    # remove rows/columns that are empty
    if 0 in row_sums or 0 in col_sums:
        # map rows with data to original rows
        row_map_nonempty = numpy.arange(original_nrows)[row_sums != 0]
        row_map_nonempty = dict(enumerate(row_map_nonempty))
        col_map_nonempty = numpy.arange(original_ncols)[col_sums != 0]
        col_map_nonempty = dict(enumerate(col_map_nonempty))

        # truncate matrix
        matrix = matrix[row_sums != 0, :][:, col_sums != 0]

        # filter row and column sums
        row_sums = row_sums[row_sums != 0]
        col_sums = col_sums[col_sums != 0]

    nrows, ncols = matrix.shape

    a = numpy.zeros((nrows, nrows), numpy.float)
    for x in range(0, nrows):
        a[x, x] = 1.0 / float(row_sums[x])

    b = numpy.zeros((ncols, ncols), numpy.float)
    for x in range(0, ncols):
        b[x, x] = 1.0 / float(col_sums[x])

    M = numpy.dot(
        a, numpy.dot(
            matrix, numpy.dot(
                b, numpy.transpose(matrix))))

    try:
        row_eigenvector = numpy.linalg.eig(M)[1][:, 1]
    except numpy.linalg.linalg.LinAlgError as msg:
        raise ValueError(msg)

    M = numpy.dot(
        b, numpy.dot(
            numpy.transpose(matrix), numpy.dot(
                a, matrix)))

    try:
        col_eigenvector = numpy.linalg.eig(M)[1][:, 1]
    except numpy.linalg.linalg.LinAlgError as msg:
        raise ValueError(msg)

    row_eigenvector = row_eigenvector.astype(numpy.float)
    col_eigenvector = col_eigenvector.astype(numpy.float)

    # insert columns ignored at the computation and give them the lowest
    # eigenvalue
    row_values = [row_eigenvector.min() - 1.0] * original_nrows
    for x, y in enumerate(row_eigenvector):
        row_values[row_map_nonempty[x]] = y
    col_values = [col_eigenvector.min() - 1.0] * original_ncols
    for x, y in enumerate(col_eigenvector):
        col_values[col_map_nonempty[x]] = y

    return row_values, col_values


def GetPermutatedMatrix(matrix,
                        map_row_new2old, map_col_new2old,
                        row_headers=None, col_headers=None):
    """return a permuted matrix. Note, that currently this is very
    inefficient, as I do not know how to do this in numpy.
    """

    nrows, ncols = matrix.shape

    result = numpy.zeros((nrows, ncols), matrix.dtype)
    for r in range(0, nrows):
        for c in range(0, ncols):
            result[r, c] = matrix[map_row_new2old[r], map_col_new2old[c]]

    if not row_headers or not col_headers:
        return result

    rows = []
    for x in map_row_new2old:
        rows.append(row_headers[x])

    cols = []
    for x in map_col_new2old:
        cols.append(col_headers[x])

    return result, rows, cols

# ---------------------------------------------------------------------


def PermuteRows(matrix):
    pass


if __name__ == "__main__":

    num_rows = 6
    num_cols = 5
    matrix = numpy.zeros((num_rows, num_cols), numpy.int)

    matrix[0, 2] = 1
    matrix[0, 3] = 1
    matrix[1, 0] = 1
    matrix[1, 1] = 1
    matrix[1, 4] = 1
    matrix[2, 1:5] = 1
    matrix[3, 2:4] = 1
    matrix[4, 0] = 1
    matrix[4, 4] = 1
    matrix[5, 0] = 1
    matrix[5, 2:5] = 1

    print("matrix=", matrix)

    row_indices, col_indices = GetIndices(matrix)

    map_row_new2old = numpy.argsort(row_indices)
    map_col_new2old = numpy.argsort(col_indices)

    print(map_row_new2old)
    print(map_col_new2old)

    print(GetPermutatedMatrix(matrix, map_row_new2old, map_col_new2old))



import os
import warnings

import matplotlib
import matplotlib.pyplot as plt
from matplotlib import _pylab_helpers

from CGATReport.Component import Component
from CGATReport import Utils

try:
    import mpld3
    HAVE_MPLD3 = True
except ImportError:
    HAVE_MPLD3 = False

try:
    import bokeh.plotting
    import bokeh.embed
    import bokeh.mpl
    import bokeh.mplexporter.exporter
    HAVE_BOKEH = True
except ImportError:
    HAVE_BOKEH = False


class MatplotlibPlugin(Component):

    capabilities = ['collect']

    def __init__(self, *args, **kwargs):
        Component.__init__(self, *args, **kwargs)

        plt.close('all')
        # matplotlib.rcdefaults()
        # set a figure size that doesn't overflow typical browser windows
        matplotlib.rcParams['figure.figsize'] = (5.5, 4.5)

    def collect(self,
                blocks,
                template_name,
                outdir,
                rstdir,
                builddir,
                srcdir,
                content,
                display_options,
                tracker_id,
                links={}):
        '''collect one or more matplotlib figures and

        1. save as png, hires-png and pdf
        2. save thumbnail
        3. insert rendering code at placeholders in output

        returns a map of place holder to placeholder text.
        '''
        fig_managers = _pylab_helpers.Gcf.get_all_fig_managers()

        map_figure2text = {}

        # determine the image formats to create
        default_format, additional_formats = Utils.getImageFormats(
            display_options)
        all_formats = [default_format, ] + additional_formats

        # create all required images
        for figman in fig_managers:

            # create all images
            figid = figman.num

            # select the correct figure
            figure = plt.figure(figid)

            # save explicit formats
            outname = "%s_%02d" % (template_name, figid)

            has_output = False

            for id, format, dpi in all_formats:

                outpath = os.path.join(outdir, '%s.%s' % (outname, format))

                # sanitize figure size for Agg.
                if figure.get_figwidth() > 32768 or figure.get_figheight() > 32768:
                    warnings.warn(
                        "figure size unexpected large {} x {}, "
                        "patching to 10x10 inches".format(
                            figure.get_figwidth(), figure.get_figheight()))
                    figure.set_size_inches(10, 10)

                try:
                    figman.canvas.figure.savefig(outpath, dpi=dpi)
                except Exception as ex:
                    s = Utils.collectExceptionAsString(
                        "exception raised while building plot '%s': %s" %
                        (outpath, str(ex)))
                    warnings.warn(s)
                    continue
                has_output = True

            if not has_output:
                warnings.warn("no output for '%s'" % outpath)
                continue

            # insert display figure
            is_html = False
            script_text = None
            if HAVE_MPLD3 and \
               Utils.PARAMS.get("report_mpl", None) == "mpld3":
                outpath = os.path.join(
                    outdir,
                    '%s.html' % (outname))

                # plt.legend()
                # write to file
                with open(outpath, "w") as outf:
                    outf.write(mpld3.fig_to_html(figure))
                is_html = True

            elif HAVE_BOKEH and \
                 Utils.PARAMS.get("report_mpl", None) == "bokeh":

                outpath = os.path.join(
                    outdir,
                    '%s.html' % (outname))
                is_html = True

                # (try to) convert to bokeh figure
                try:
                    # pd_job - use pandas object
                    # xkcd - use xkcd style
                    renderer = bokeh.mpl.BokehRenderer(pd_obj=True,
                                                       xkcd=False)
                    exporter = bokeh.mplexporter.exporter.Exporter(renderer)
                    exporter.run(figure)
                    bpl = renderer.fig

                except NotImplementedError:
                    # fall back to matplotlib
                    is_html = False

                if is_html:
                    bokeh_id = bpl._id
                    res = bokeh.resources.CDN
                    script_path = os.path.join(
                        '_static/report_directive/',
                        "%s.js" % bokeh_id)

                    # get js figure and html snippet
                    js_text, script_text = bokeh.embed.autoload_static(
                        bpl, res, script_path)

                    with open(script_path, "w") as outf:
                        outf.write(js_text)

                    with open(outpath, "w") as outf:
                        outf.write(script_text)

            # create the text element
            rst_output = Utils.buildRstWithImage(
                outname,
                outdir,
                rstdir,
                builddir,
                srcdir,
                additional_formats,
                tracker_id,
                links,
                display_options,
                default_format,
                is_html=is_html,
                text=script_text)

            map_figure2text["#$mpl %i$#" % figid] = rst_output
            
        return map_figure2text



# -*- coding: utf-8 -*-
"""
    sphinx.ext.warnings
    ~~~~~~~~~~~~~~~~~~~

    Collect all warnings

:copyright: Copyright 2007-2009 by the Andreas Heger, see AUTHORS.
:license: BSD, see LICENSE for details.
"""

import logging
from docutils import nodes

from sphinx.locale import _
from sphinx.environment import NoUri
from sphinx.util.compat import Directive
from docutils.parsers.rst.directives.admonitions import BaseAdmonition


class cgatreportwarning_node(nodes.warning, nodes.Element):
    pass


class cgatreportwarninglist(nodes.General, nodes.Element):
    pass


class CGATReportWarning(BaseAdmonition):

    node_class = cgatreportwarning_node

    # accept error as an optional argument
    optional_arguments = 1

    def run(self):

        r = BaseAdmonition.run(self)

        if len(self.arguments) > 0:
            warningclass = self.arguments[0]
        else:
            warningclass = "generic"

        env = self.state.document.settings.env
        if not hasattr(env, 'cgatreportwarning_all_cgatreportwarnings'):
            env.cgatreportwarning_all_cgatreportwarnings = []

        env.cgatreportwarning_all_cgatreportwarnings.append({
            'docname': env.docname,
            'lineno': self.lineno,
            'cgatreportwarning': r[0].deepcopy(),
            'warningclass': warningclass,
        })

        logging.warn("CGATReport-Warning: %s" % warningclass)

        return r


class CGATReportWarningList(Directive):

    """
    A list of all cgatreportwarning entries.
    """

    has_content = False
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {}

    def run(self):
        # Simply insert an empty cgatreportwarninglist node which will
        # be replaced later when process_cgatreportwarning_nodes is
        # called
        return [cgatreportwarninglist('')]


def process_cgatreportwarning_nodes(app, doctree, fromdocname):
    if not app.config['cgatreport_show_warnings']:
        for node in doctree.traverse(cgatreportwarning_node):
            node.parent.remove(node)

    # Replace all cgatreportwarninglist nodes with a list of the
    # collected cgatreportwarnings.  Augment each cgatreportwarning
    # with a backlink to the original location.
    env = app.builder.env

    if not hasattr(env, 'cgatreportwarning_all_cgatreportwarnings'):
        env.cgatreportwarning_all_cgatreportwarnings = []

    for node in doctree.traverse(cgatreportwarninglist):
        if not app.config['cgatreport_show_warnings']:
            node.replace_self([])
            continue

        content = []
        nwarnings = 0

        para = nodes.paragraph()
        para += nodes.Text("There are %i warnings" %
                           len(env.cgatreportwarning_all_cgatreportwarnings))
        content.append(para)

        for cgatreportwarning_info in env.cgatreportwarning_all_cgatreportwarnings:

            para = nodes.paragraph()

            filename = env.doc2path(
                cgatreportwarning_info['docname'], base=None)

            nwarnings += 1
            location_str = '%s:%d ' % (
                filename, cgatreportwarning_info['lineno'])
            try:
                description_str = cgatreportwarning_info['warningclass']
            except KeyError:
                description_str = "unknown"

            # Create a reference
            newnode = nodes.reference('', '')
            innernode = nodes.emphasis(_(location_str), _(location_str))
            newnode['refdocname'] = cgatreportwarning_info['docname']

            try:
                newnode['refuri'] = app.builder.get_relative_uri(
                    fromdocname, cgatreportwarning_info['docname'])
            except NoUri:
                # ignore if no URI can be determined, e.g. for LaTeX output
                pass
            newnode.append(innernode)

            para += newnode
            para += nodes.Text(description_str, description_str)
            para += nodes.Text("\n", "\n")

            # (Recursively) resolve references in the cgatreportwarning content
            cgatreportwarning_entry = cgatreportwarning_info[
                'cgatreportwarning']
            env.resolve_references(cgatreportwarning_entry,
                                   cgatreportwarning_info['docname'],
                                   app.builder)

            content.append(para)

        node.replace_self(content)


def purge_cgatreportwarnings(app, env, docname):
    if not hasattr(env, 'cgatreportwarning_all_cgatreportwarnings'):
        return
    env.cgatreportwarning_all_cgatreportwarnings = \
        [cgatreportwarning for cgatreportwarning in
         env.cgatreportwarning_all_cgatreportwarnings
         if cgatreportwarning['docname'] != docname]


def visit_cgatreportwarning_node(self, node):
    self.visit_admonition(node)


def depart_cgatreportwarning_node(self, node):
    self.depart_admonition(node)


def setup(app):
    app.add_config_value('cgatreport_show_warnings', True, False)

    app.add_node(cgatreportwarninglist)
    app.add_node(cgatreportwarning_node,
                 html=(visit_cgatreportwarning_node,
                       depart_cgatreportwarning_node),
                 latex=(
                     visit_cgatreportwarning_node,
                     depart_cgatreportwarning_node),
                 text=(visit_cgatreportwarning_node,
                       depart_cgatreportwarning_node))

    app.add_directive('warning', CGATReportWarning)
    app.add_directive('warninglist', CGATReportWarningList)
    app.connect('doctree-resolved', process_cgatreportwarning_nodes)
    app.connect('env-purge-doc', purge_cgatreportwarnings)



'''Script to fill csvdb database with example data.'''

import random
import os
from subprocess import Popen

cmd = 'sqlite3 csvdb "%s" '


def e(stmt):
    p = Popen( cmd % stmt, shell=True)
    os.waitpid(p.pid, 0)

e("CREATE TABLE experiment1_data "
  "(gene_id TEXT, function TEXT, expression INT)")

for x in range(0, 200, 2):
    e( "INSERT INTO %s VALUES ('gene%2i', '%s', %f)" %
       ("experiment1_data", x, "housekeeping", random.gauss(41, 5)))
    e( "INSERT INTO %s VALUES ('gene%2i', '%s', %f)" %
       ("experiment1_data", x+1, "regulation", random.gauss(11, 5)))

e("CREATE TABLE experiment2_data "
  "(gene_id TEXT, function TEXT, expression INT)")

for x in range(0, 200, 2):
    e( "INSERT INTO %s VALUES ('gene%2i', '%s', %f)" %
       ("experiment2_data", x, "housekeeping", random.gauss(40, 5)))
    e( "INSERT INTO %s VALUES ('gene%2i', '%s', %f)" %
       ("experiment2_data", x+1, "regulation", random.gauss(10, 5)))





"""A special directive for including a matplotlib plot.

Given a path to a .py file, it includes the source code inline, then:

- On HTML, will include a .png with a link to a high-res .png.

- On LaTeX, will include a .pdf

This directive supports all of the options of the `image` directive,
except for `target` (since plot will add its own target).

"""

import os
import hashlib
import re
import logging
import collections

from docutils.parsers.rst import directives
from docutils.parsers.rst import Directive

from CGATReport import Config, Dispatcher, Utils, Cache, Component
from CGATReport.ResultBlock import ResultBlocks

CGATREPORT_DEBUG = "CGATREPORT_DEBUG" in os.environ

TEMPLATE_TEXT = """
.. only:: html

   %(url_template)s

"""


def out_of_date(original, derived):
    """
    Returns True if derivative is out-of-date wrt original,
    both of which are full file paths.
    """
    return (not os.path.exists(derived)
            or os.stat(derived).st_mtime < os.stat(original).st_mtime)


def run(arguments,
        options,
        lineno,
        content,
        state_machine=None,
        document=None,
        srcdir=None,
        builddir=None):
    """process:report: directive.

    *srdir* - top level directory of rst documents
    *builddir* - build directory
    """

    tag = "%s:%i" % (str(document), lineno)

    logger = Component.get_logger()

    logger.debug("report_directive.run: profile: started: rst: %s" % tag)

    # sort out the paths
    # reference is used for time-stamping
    tracker_name = directives.uri(arguments[0])

    (basedir, fname, basename, ext, outdir,
     codename, notebookname) = Utils.build_paths(tracker_name)

    # get the directory of the rst file
    # state_machine.document.attributes['source'])
    rstdir, rstfile = os.path.split(document)
    # root of document tree
    if srcdir is None:
        srcdir = setup.srcdir

    # build directory
    if builddir is None:
        builddir = setup.builddir

    # remove symbolic links
    srcdir, builddir, rstdir = [
        os.path.abspath(os.path.realpath(x)) for x in (srcdir, builddir, rstdir)]

    # there are three directories:
    # builddir = directory where document is built in
    #            (usually _build/html or similar)
    # rstdir   = directory where rst sources are located
    # srcdir   = directory from which the build process is started

    # path to root relative to rst
    rst2srcdir = os.path.join(os.path.relpath(srcdir, start=rstdir), outdir)

    # path to root relative to rst
    rst2builddir = os.path.join(
        os.path.relpath(builddir, start=rstdir), outdir)

    # path relative to source (for images)
    root2builddir = os.path.join(
        os.path.relpath(builddir, start=srcdir), outdir)

    logger.debug(
        "report_directive.run: arguments=%s, options=%s, lineno=%s, "
        "content=%s, document=%s" %
        (str(arguments),
         str(options),
         str(lineno),
         str(content),
         str(document)))
    logger.debug(
        "report_directive.run: plotdir=%s, basename=%s, ext=%s, "
        "fname=%s, rstdir=%s, srcdir=%s, builddir=%s" %
        (tracker_name, basename, ext, fname, rstdir, srcdir, builddir))
    logger.debug(
        "report_directive.run: tracker_name=%s, basedir=%s, "
        "rst2src=%s, root2build=%s, outdir=%s, codename=%s" %
        (tracker_name, basedir, rst2srcdir, rst2builddir, outdir, codename))

    # try to create. If several processes try to create it,
    # testing with `if` will not work.
    try:
        os.makedirs(outdir)
    except OSError as msg:
        pass

    if not os.path.exists(outdir):
        raise OSError("could not create directory %s: %s" % (outdir, msg))

    ########################################################
    # collect options
    # replace placedholders
    try:
        options = Utils.updateOptions(options)
    except ValueError as msg:
        logger.warn("failure while updating options: %s" % msg)

    logger.debug("report_directive.run: options=%s" % (str(options),))

    transformer_names = []
    renderer_name = None

    layout = options.get("layout", "column")
    option_map = Component.getOptionMap()
    renderer_options = Utils.selectAndDeleteOptions(
        options, option_map["render"])
    transformer_options = Utils.selectAndDeleteOptions(
        options, option_map["transform"])
    dispatcher_options = Utils.selectAndDeleteOptions(
        options, option_map["dispatch"])
    tracker_options = Utils.selectAndDeleteOptions(
        options, option_map["tracker"])
    display_options = Utils.get_default_display_options()
    display_options.update(Utils.selectAndDeleteOptions(
        options, option_map["display"]))

    logger.debug("report_directive.run: renderer options: %s" %
                  str(renderer_options))
    logger.debug("report_directive.run: transformer options: %s" %
                  str(transformer_options))
    logger.debug("report_directive.run: dispatcher options: %s" %
                  str(dispatcher_options))
    logger.debug("report_directive.run: tracker options: %s" %
                  str(tracker_options))
    logger.debug("report_directive.run: display options: %s" %
                  str(display_options))

    if "transform" in display_options:
        transformer_names = display_options["transform"].split(",")
        del display_options["transform"]

    if "render" in display_options:
        renderer_name = display_options["render"]
        del display_options["render"]

    ########################################################
    # check for missing files
    if renderer_name is not None:

        options_key = str(renderer_options) +\
            str(transformer_options) +\
            str(dispatcher_options) +\
            str(tracker_options) +\
            str(transformer_names) +\
            re.sub("\s", "", "".join(content))
                
        options_hash = hashlib.md5(options_key.encode()).hexdigest()

        template_name = Utils.quote_filename(
            Config.SEPARATOR.join((tracker_name, renderer_name,
                                   options_hash)))
        filename_text = os.path.join(outdir, "%s.txt" % (template_name))
        rstname = os.path.basename(filename_text)
        notebookname += options_hash

        logger.debug("report_directive.run: options_hash=%s" % options_hash)

        ###########################################################
        # check for existing files
        # update strategy does not use file stamps, but checks
        # for presence/absence of text element and if all figures
        # mentioned in the text element are present
        ###########################################################
        queries = [re.compile("%s/(\S+.%s)" %
                              (root2builddir, suffix))
                   for suffix in ("png", "pdf", "svg")]

        logger.debug("report_directive.run: checking for changed files.")

        # check if text element exists
        if os.path.exists(filename_text):

            with open(filename_text, "r") as inf:
                lines = [x[:-1] for x in inf]
            filenames = []

            # check if all figures are present
            for line in lines:
                for query in queries:
                    x = query.search(line)
                    if x:
                        filenames.extend(list(x.groups()))

            filenames = [os.path.join(outdir, x) for x in filenames]

            logger.debug(
                "report_directive.run: %s: checking for %s" %
                (tag, str(filenames)))
            for filename in filenames:
                if not os.path.exists(filename):
                    logger.info(
                        "report_directive.run: %s: redo: file %s is missing" %
                        (tag, filename))
                    break
            else:
                logger.info(
                    "report_directive.run: %s: noredo: all files are present" %
                    tag)
                # all is present - save text and return
                if lines and state_machine:
                    state_machine.insert_input(
                        lines, state_machine.input_lines.source(0))
                return []
        else:
            logger.debug(
                "report_directive.run: %s: no check performed: %s missing" %
                (tag, str(filename_text)))
    else:
        template_name = ""
        filename_text = None

    ##########################################################
    # Initialize collectors
    collectors = []
    for collector in list(Component.getPlugins("collect").values()):
        collectors.append(collector())

    ##########################################################
    # instantiate tracker, dispatcher, renderer and transformers
    # and collect output
    ###########################################################
    try:
        ########################################################
        # find the tracker
        logger.debug(
            "report_directive.run: collecting tracker %s with options %s " %
            (tracker_name, tracker_options))
        code, tracker, tracker_path = Utils.makeTracker(
            tracker_name, (), tracker_options)
        if not tracker:
            logger.error(
                "report_directive.run: no tracker - no output from %s " %
                str(document))
            raise ValueError("tracker `%s` not found" % tracker_name)

        logger.debug(
            "report_directive.run: collected tracker %s" % tracker_name)

        tracker_id = Cache.tracker2key(tracker)

        ########################################################
        # determine the transformer
        logger.debug("report_directive.run: creating transformers")

        transformers = Utils.getTransformers(
            transformer_names, transformer_options)

        ########################################################
        # determine the renderer
        logger.debug("report_directive.run: creating renderer.")

        if renderer_name is None:
            logger.error(
                "report_directive.run: no renderer - no output from %s" %
                str(document))
            raise ValueError("the report directive requires a renderer")

        renderer = Utils.getRenderer(renderer_name, renderer_options)

        try:
            renderer.set_paths(rstdir, srcdir, builddir)
            renderer.set_display_options(display_options)
        except AttributeError:
            # User renderers will not have these methods
            pass

        ########################################################
        # create and call dispatcher
        logger.debug("report_directive.run: creating dispatcher")

        dispatcher = Dispatcher.Dispatcher(tracker,
                                           renderer,
                                           transformers)

        # add the tracker options
        dispatcher_options.update(tracker_options)

        blocks = dispatcher(**dispatcher_options)

        if blocks is None:
            blocks = ResultBlocks(ResultBlocks(
                Utils.buildWarning(
                    "NoData",
                    "tracker %s returned no Data" % str(tracker))))
            code = None
            tracker_id = None

    except:

        logger.warn(
            "report_directive.run: exception caught at %s:%i - see document" %
            (str(document), lineno))

        blocks = ResultBlocks(ResultBlocks(
            Utils.buildException("invocation")))
        code = None
        tracker_id = None

    logger.debug(
        "report_directive.run: profile: started: collecting: %s" % tag)

    ########################################################
    # write code output
    linked_codename = re.sub("\\\\", "/", os.path.join(rst2builddir, codename))
    if code and basedir != outdir:
        with open(os.path.join(outdir, codename), "w") as outfile:
            for line in code:
                outfile.write(line)

    ########################################################
    # write notebook snippet
    linked_notebookname = re.sub(
        "\\\\", "/", os.path.join(rst2builddir, notebookname))

    if basedir != outdir and tracker_id is not None:
        with open(os.path.join(outdir, notebookname), "w") as outfile:
            Utils.writeNoteBookEntry(outfile,
                                     renderer=renderer_name,
                                     tracker=tracker_name,
                                     transformers=transformer_names,
                                     tracker_path=tracker_path,
                                     options=renderer_options.items() +
                                     tracker_options.items() +
                                     transformer_options.items())

    if filename_text is not None:
        linked_rstname = re.sub(
            "\\\\", "/", os.path.join(rst2builddir, rstname))
    else:
        linked_rstname = None

    ###########################################################
    # collect images
    ###########################################################
    map_figure2text = {}
    links = {'code_url': linked_codename,
             'rst_url': linked_rstname,
             'notebook_url': linked_notebookname}
    try:
        for collector in collectors:
            map_figure2text.update(collector.collect(
                blocks,
                template_name,
                outdir,
                rstdir,
                builddir,
                srcdir,
                content,
                display_options,
                tracker_id,
                links=links))
    except:

        logger.warn("report_directive.run: exception caught while "
                     "collecting with %s at %s:%i - see document" %
                     (collector, str(document), lineno))
        blocks = ResultBlocks(ResultBlocks(
            Utils.buildException("collection")))
        code = None
        tracker_id = None

    ###########################################################
    # replace place holders or add text
    ###########################################################
    # add default for text-only output
    requested_urls = Utils.asList(Utils.PARAMS["report_urls"])

    urls = []
    if "code" in requested_urls:
        urls.append(":download:`code <%(code_url)s>`" % links)

    if "notebook" in requested_urls:
        urls.append(":download:`nb <%(notebook_url)s>`" % links)
    
    map_figure2text["default-prefix"] = ""
    map_figure2text["default-suffix"] = ""

    if urls and "no-links" not in display_options:
        url_template = "[{}]".format(" ".join(urls))
    else:
        url_template = ""

    map_figure2text["default-prefix"] = TEMPLATE_TEXT % locals()

    blocks.updatePlaceholders(map_figure2text)

    # render the output taking into account the layout
    lines = Utils.layoutBlocks(blocks, layout)
    lines.append("")

    # add caption
    if content and "no-caption" not in display_options:
        lines.extend(['::', ''])
        lines.extend(['    %s' % row.strip() for row in content])
        lines.append("")

    # output rst text for this renderer
    if filename_text:
        outfile = open(filename_text, "w")
        outfile.write("\n".join(lines))
        outfile.close()

    if CGATREPORT_DEBUG:
        for x, l in enumerate(lines):
            print("%5i %s" % (x, l))

    if len(lines) and state_machine:
        state_machine.insert_input(
            lines, state_machine.input_lines.source(0))

    logger.debug(
        "report_directive.run: profile: finished: collecting: %s" % tag)
    logger.debug(
        "report_directive.run: profile: finished: rst: %s:%i" %
        (str(document), lineno))

    return []


class report_directive(Directive):
    required_arguments = 1
    optional_arguments = 0
    has_content = True
    final_argument_whitespace = True

    # build option spec
    option_spec = Component.getOptionSpec()

    def run(self):
        document = self.state.document.current_source
        logger = Component.get_logger()
        logger.info("report_directive: starting: %s:%i" %
                     (str(document), self.lineno))

        return run(self.arguments,
                   self.options,
                   self.lineno,
                   self.content,
                   self.state_machine,
                   document)


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir
    setup.srcdir = app.srcdir
    setup.builddir = os.getcwd()
    app.add_directive('report', report_directive)

    # update global parameters in Utils module.
    PARAMS = Utils.get_parameters()
    app.add_config_value('PARAMS', collections.defaultdict(), 'env')

    setup.logger = Component.get_logger()

    # return {'parallel_read_safe': True}

directives.register_directive('report', report_directive)




from CGATReport.Component import *
from CGATReport import Config, Utils

import os
import re

try:
    from rpy2.robjects import r as R
    import rpy2.robjects as ro
    import rpy2.robjects.numpy2ri
    import rpy2.rinterface
except ImportError:
    R = None

import matplotlib.image as image

import warnings


class RPlotPlugin(Component):

    capabilities = ['collect']

    def __init__(self, *args, **kwargs):
        Component.__init__(self, *args, **kwargs)

    def collect(self,
                blocks,
                template_name,
                outdir,
                rstdir,
                builddir,
                srcdir,
                content,
                display_options,
                tracker_id,
                links={}):
        '''collect one or more R figures.

        Plots are collected from all active devices.
        Plots are also collected from result-blocks
        containing a 'ggplot' attribute.

        1. save as png, hires-png and pdf
        2. save thumbnail
        3. insert rendering code at placeholders in output

        returns a map of place holder to placeholder text.
        '''
        # disable plotting if no rpy installed
        if R is None:
            return {}

        map_figure2text = {}

        # determine the image formats to create
        default_format, additional_formats = Utils.getImageFormats(
            display_options)
        all_formats = [default_format, ] + additional_formats
        image_options = Utils.getImageOptions(display_options)

        ##########################################
        ##########################################
        ##########################################
        # iterate over devices
        devices = R["dev.list"]()
        try:
            maxid = max(R["dev.list"]())
        except TypeError:
            maxid = 0

        for figid in range(2, maxid + 1):

            for id, format, dpi in all_formats:

                R["dev.set"](figid)

                outname = "%s_%02d" % (template_name, figid)
                outpath = os.path.join(outdir, '%s.%s' % (outname, format))

                if format.endswith("png"):
                    # for busy images there is a problem with figure margins
                    # simply increase dpi until it works.
                    R["dev.set"](figid)

                    width = height = 480 * dpi / 80
                    x = 0
                    while 1:
                        try:
                            R["dev.copy"](device=R.png,
                                          filename=outpath,
                                          res=dpi,
                                          width=width,
                                          height=height)
                            R["dev.off"]()
                        except rpy2.rinterface.RRuntimeError:
                            width *= 2
                            height *= 2
                            if x < 5:
                                continue
                        break

                elif format.endswith("svg"):
                    R["dev.copy"](device=R.svg,
                                  filename=outpath)
                    R["dev.off"]()

                elif format.endswith("eps"):
                    R["dev.copy"](device=R.postscript,
                                  paper='special',
                                  width=6,
                                  height=6,
                                  file=outpath,
                                  onefile=True)
                    R["dev.off"]()
                elif format.endswith("pdf"):
                    R["dev.copy"](device=R.pdf,
                                  paper='special',
                                  width=6,
                                  height=6,
                                  file=outpath,
                                  onefile=True)
                    R["dev.off"]()
                else:
                    raise ValueError("format '%s' not supported" % format)

                if not os.path.exists(outpath):
                    continue
                    # raise ValueError("rendering problem: image file was not be created: %s" % outpath)

                if format == 'png':
                    thumbdir = os.path.join(outdir, 'thumbnails')
                    try:
                        os.makedirs(thumbdir)
                    except OSError:
                        pass
                    thumbfile = str('%s.png' % os.path.join(thumbdir, outname))
                    captionfile = str(
                        '%s.txt' % os.path.join(thumbdir, outname))
                    if not os.path.exists(thumbfile):
                        # thumbnail only available in matplotlib >= 0.98.4
                        try:
                            figthumb = image.thumbnail(
                                str(outpath), str(thumbfile), scale=0.3)
                        except AttributeError:
                            pass
                    outfile = open(captionfile, "w")
                    outfile.write("\n".join(content) + "\n")
                    outfile.close()

                R["dev.off"](figid)

            # create the text element
            rst_output = Utils.buildRstWithImage(outname,
                                                 outdir,
                                                 rstdir,
                                                 builddir,
                                                 srcdir,
                                                 additional_formats,
                                                 tracker_id,
                                                 links,
                                                 display_options)

            map_figure2text["#$rpl %i$#" % figid] = rst_output

        figid = maxid
        ##########################################
        ##########################################
        ##########################################
        # iterate over ggplot plots
        for xblocks in blocks:
            for block in xblocks:
                if not hasattr(block, "rggplot"):
                    continue
                pp = block.rggplot
                figname = block.figname

                outname = "%s_%s" % (template_name, figname)

                for id, format, dpi in all_formats:
                    outpath = os.path.join(outdir, '%s.%s' % (outname, format))

                    try:
                        R.ggsave(outpath, plot=pp, dpi=dpi)
                    except rpy2.rinterface.RRuntimeError, msg:
                        raise

                    # width, height = 3 * dpi, 3 * dpi
                    # if format.endswith("png"):
                    #     R.png(outpath,
                    #            width = width,
                    #            height = height)
                    # elif format.endswith("svg"):
                    #     R.svg(outpath)
                    # elif format.endswith("eps"):
                    #     R.postscript(outpath)
                    # elif format.endswith("pdf"):
                    #     R.pdf(outpath)
                    # R.plot(pp)
                    # R["dev.off"]()

                # create the text element
                rst_output = Utils.buildRstWithImage(outname,
                                                     outdir,
                                                     rstdir,
                                                     builddir,
                                                     srcdir,
                                                     additional_formats,
                                                     tracker_id,
                                                     links,
                                                     display_options)

                map_figure2text["#$ggplot %s$#" % figname] = rst_output

        return map_figure2text



import sys

import matplotlib
matplotlib.use(sys.argv[1])
import pandas
import numpy
import matplotlib.pyplot as plt
from matplotlib import _pylab_helpers
import timeit

datapoints = int(sys.argv[2])
figures = 10

backend = matplotlib.get_backend()

for figure in range(figures):
    df = pandas.DataFrame(numpy.random.rand(datapoints, 2),
                          columns=['a', 'b'])

    df.plot.scatter(x='a', y='b')

def write_png(dpi):

    fig_managers = _pylab_helpers.Gcf.get_all_fig_managers()

    # create all required images
    for figman in fig_managers:

        # create all images
        figid = figman.num

        # select the correct figure
        figure = plt.figure(figid)

        outpath = "test_writing_%i.jpg" % figid
        figman.canvas.figure.savefig(outpath, dpi=dpi)


if __name__ == '__main__':
    import timeit
    for dpi in [50, 100, 200, 300, 400]:
        print "\t".join(map(
            str,
            (backend, dpi, timeit.timeit(
            "write_png(dpi={})".format(dpi),
            setup="from __main__ import write_png",
                number=3))))
